"main testing module"
from django.test import TestCase

def data():
    "Add data to the system to test against."
    # pylint: disable=no-member
    # Add the following items:
    from ..tools import models
    entries = ['Alpha', 'Bravo', 'Charlie', 'Delta',
               'A1', 'C1', 'D1', 'D2',
               'C1-A', 'D2-1', 'D2-2']

    tmp = dict()
    for index, entry in enumerate(entries):
        index += 1
        tmp[entry] = models.Item.objects.create(id=index, name=entry)

    # Now create relationships, A1 is under Alpha, C1 under Charlie and D1 & D2
    # under Delta, C1-A under C1, etc.
    pairs = [[tmp['Alpha'], tmp['A1']],
             [tmp['Charlie'], tmp['C1']],
             [tmp['Delta'], tmp['D1']],
             [tmp['Delta'], tmp['D2']],
             [tmp['C1'], tmp['C1-A']],
             [tmp['D2'], tmp['D2-1']],
             [tmp['D2'], tmp['D2-2']]]

    for one, two in pairs:
        models.Link.objects.create(item_one=one, item_two=two)



# Create your tests here.
class MainTestCase(TestCase):
    "main test case"
    # pylint: disable=missing-docstring, no-member, invalid-name
    @classmethod
    def setUpClass(cls):
        returns = super(MainTestCase, cls).setUpClass()
        data()
        return returns

    def setUp(self):
        self.expected_4 = {4:{7:'$',
                              8:{10:'$',
                                 11:'$'}},}
        TestCase.setUp(self)

    def test_010_delta_id(self):
        from .. import models
        expected = 4
        actually = models.Item.objects.get(name='Delta').id
        self.assertEqual(expected, actually)

    def test_020_get_linked_ids(self):
        from ..tools import relations
        expected = [7, 8]
        actually = list(relations.get_linked_ids(4))
        self.assertEqual(expected, actually)

    def test_021_get_linked_ids_upwards(self):
        from ..tools import relations
        expected = [1]
        actually = list(relations.get_linked_ids(5, upwards=True))
        self.assertEqual(expected, actually)

    def test_030_get_work(self):
        from ..tools import relations
        temp = {'Alpha':{'A1':None},
                'Bravo':None,
                'Charlie':{'C1':{'C1-A':None},},
                'Delta':{'D1':None,
                         'D2':{'D2-1':None,
                               'D2-2':'XYZ'}}}
        expected = 'XYZ'
        provided = ['Delta', 'D2', 'D2-2']
        actually = relations.get_work(temp, provided)
        self.assertEqual(expected, actually)

    def test_040_get_all_leaves(self):
        "Test if it gets all leaves"
        from ..tools import relations
        provided = 4
        expected = self.expected_4
        actually = relations.get_nodes_recursive_from(provided)
        self.assertEqual(expected, actually)

    def test_050_get_all_leaves_with_infinity(self):
        "Test if it gets all leaves"
        # As we are adding a link, this test will also fail if the lru_cache
        # is not cleared (since we already queries with the same parameters).
        from ..tools import relations, models
        item_one = models.Item.objects.get(name='D2-2')
        item_two = models.Item.objects.get(name='Delta')
        link = models.Link.objects.create(item_one=item_one,
                                          item_two=item_two)
        self.expected_4[4][8][11] = {4:'∞'}

        provided = 4
        expected = self.expected_4
        actually = relations.get_nodes_recursive_from(provided)
        self.assertEqual(expected, actually)
        link.delete()

    def test_060_get_leaves_at_end(self):
        "Test if it gets all leaves"
        from ..tools import relations
        provided = 10
        expected = {10:'$'}
        actually = relations.get_nodes_recursive_from(provided)
        self.assertEqual(expected, actually)

    def test_070_relations_exception(self):
        "Test if throws an exception if the item id does not exist."
        from ..tools import relations
        provided = 20
        self.assertRaises(relations.RelationsError,
                          relations.get_nodes_recursive_from, provided)

    def test_080_get_leaves_by_item(self):
        "Test if the model integations work"
        from ..tools import models
        provided = models.Item.objects.get(id=4)
        expected = self.expected_4
        actually = provided.get_linked()
        self.assertEqual(expected, actually)


    def test_085_add_link_via_item(self):
        "Test if it gets all leaves"
        from ..tools import relations, models
        item = models.Item.objects.get(name='D2-2')
        to_link = models.Item.objects.get(name='Delta')
        link = item.add_link(to_link)

        self.expected_4[4][8][11] = {4:'∞'}

        provided = 4
        expected = self.expected_4
        actually = relations.get_nodes_recursive_from(provided)
        self.assertEqual(expected, actually)
        link.delete()

    def test_090_get_leaves_by_item_upwards(self):
        "Test if the model integations work"
        from ..tools import models
        provided = models.Item.objects.get(id=5)
        expected = {5: {1: '^'}}
        actually = provided.get_linked(upwards=True)
        self.assertEqual(expected, actually)

    def test_100_print_name(self):
        "Test if the models string are as expected"
        from ..tools import models
        item = models.Item.objects.get(id=5)
        self.assertEqual('A1', str(item))

        link = item.linked_as_two.all()[0]

        self.assertEqual('Alpha | A1', str(link))
        relation_name = 'Relation Name'

        relation = models.Relation.objects.create(name=relation_name)
        self.assertEqual(relation_name, str(relation))

        link.relation = relation
        link.save()

        self.assertEqual('Alpha | A1' + ' # ' + relation_name, str(link))

    def test_110_view(self):
        "Test if the view returns something sensible."
        # The view is not expected to be used in anything resembling production
        response = self.client.get('/')
        expected = [[1, 5, None], [3, 6, None], [4, 7, None], [4, 8, None],
                    [6, 9, None], [8, 10, None], [8, 11, None]]
        self.assertEqual(expected, response.json()['response'])


if __name__ == '__main__': # pragma: no cover
    # pylint: disable=wrong-import-position
    import django
    django.setup()
    django.core.management.call_command('test')



"""
Django Linked Items Models
"""
from django.db import models
from ..tools import relations

#pylint: disable=no-member
class Item(models.Model):
    "The base item to be linked."
    name = models.CharField(max_length=128)

    def get_linked(self, upwards=False):
        "Return relations."
        # The return compound dictionary has keys that are the item id's.
        return relations.get_nodes_recursive_from(self.id, upwards)

    def add_link(self, item, relation=None):
        "Add a link"
        return Link.objects.create(item_one=self, item_two=item,
                                   relation=relation)

    def __str__(self):
        return self.name


class Relation(models.Model):
    "The relationship between the item."
    name = models.CharField(max_length=128)

    def __str__(self):
        return self.name


class Link(models.Model):
    "Link two items together"
    class Meta: # pylint: disable=missing-docstring, too-few-public-methods
        unique_together = ('item_one', 'item_two', 'relation')

    item_one = models.ForeignKey(Item, related_name='linked_as_one')
    item_two = models.ForeignKey(Item, related_name='linked_as_two')
    relation = models.ForeignKey(Relation, related_name='link',
                                 null=True, blank=True)

    def __str__(self):
        tmp = [str(self.item_one), ' | ', str(self.item_two)]
        if not self.relation is None:
            tmp.append(' # ' + str(self.relation))
        return ''.join(tmp)

    def save(self, *args, **kwargs):
        return models.Model.save(self, *args, **kwargs)




# -*- coding: utf-8 -*-
# pylint: disable=invalid-name, missing-docstring, line-too-long
# Generated by Django 1.9.7 on 2016-06-09 11:05
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Item',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
            ],
        ),
        migrations.CreateModel(
            name='Link',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('item_one', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='linked_as_one', to='django_linked_items.Item')),
                ('item_two', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='linked_as_two', to='django_linked_items.Item')),
            ],
        ),
        migrations.CreateModel(
            name='Relation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
            ],
        ),
        migrations.AddField(
            model_name='link',
            name='relation',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='link', to='django_linked_items.Relation'),
        ),
        migrations.AlterUniqueTogether(
            name='link',
            unique_together=set([('item_one', 'item_two', 'relation')]),
        ),
    ]



"""
Lookup ids
"""
from functools import lru_cache
# pylint: disable=no-member

class RelationsError(ValueError):
    "A relation value error"
    pass

@lru_cache()
def get_linked_ids(item_id, upwards=False):
    "get the ids of the items that are linked as item two."
    from ..tools import models
    query = models.Item.objects.filter(id=item_id)
    if not query.exists():
        raise RelationsError("Item id '%s' does not exist." % item_id)

    values = list()
    if upwards:
        values.append('linked_as_two__item_one')
    else:
        values.append('linked_as_one__item_two')

    ids = query.values_list(*values, flat=True)
    ids = list(ids)
    while None in ids:
        ids.remove(None)

    return ids

def get_work(temp, index):
    "Go through the index and return the last indexed one."
    path = index[::]
    work = temp
    while len(path) > 0:
        work = work[path.pop(0)]

    return work


def _process_todo(todo, done, work, upwards=False):
    "process the todo list."
    symbol_infinity = '∞'
    # The symbols ^ and $ are inspired from Regular Expressions where they
    # represent the beginning and end of a line.
    if upwards:
        symbol_endpoint = '^'
    else:
        symbol_endpoint = '$'

    index = todo.pop(0)
    item_id = index[-1]
    done.add(item_id)

    links = get_linked_ids(item_id, upwards)

    if len(links) == 0:
        # This link is an end point, e.g. no further connections.
        work = get_work(work, index[:-1])
        work[item_id] = symbol_endpoint
    else:
        # This link has connected nodes, process each node and append the new
        # nodes to the todo list.
        work = get_work(work, index)
        for link in links:
            if link in done:
                # Already saw this item earlier, thus it is an infinite loop.
                work[link] = symbol_infinity
                continue

            work[link] = dict()
            append_index = index[::] + [link]
            todo.append(append_index)

@lru_cache()
def get_nodes_recursive_from(base_id, upwards=False):
    "Return connected nodes recursively starting from base_id."
    # But using a non-recursive process to prevent max recursion error.
    # By default we go forwards, i.e. return nodes that have this node as a
    # parent, specifying upwards=True, makes it go the other way, i.e. nodes
    # that have this node as a child.
    work = {}
    done = set()
    todo = list() # Todo is a list of index lists, see 'get_work' function.

    # Set up the first working item
    todo.append([base_id,])
    work[base_id] = dict()

    while len(todo) > 0:
        # The _process_todo will add items to the todo list when it finds
        # related nodes. This is the way we can recursively walk through the
        # dependencies without resorting to recursive function calling.
        _process_todo(todo, done, work, upwards)

    return work





"""
This module is imported on app ready (see __init__).
"""

from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver
from . import models
from .tools import relations


@receiver((post_save, post_delete), sender=models.Link)
def lru_cache_clear(**kwargs): # pylint: disable=unused-argument
    "Clear the lru cache."
    to_clear = [relations.get_linked_ids, relations.get_nodes_recursive_from]

    for lru_chached in to_clear:
        lru_chached.cache_clear()



"""
Project Views
"""

from django.http import JsonResponse
from ..models import Link

# pylint: disable=no-member, unused-argument
def view(request):
    "Example view"
    links = Link.objects.all().values_list('item_one__id',
                                           'item_two__id',
                                           'relation__name')
    return JsonResponse({'response':list(links)})




"""
PyPi Setup file.
"""
import os
# pylint: disable=no-name-in-module, import-error
from setuptools import setup, find_packages

NAME = 'django-linked-items'
DESCRIPTION = 'Linked Items'
VERSION = '1.0.0.3'
AUTHOR = 'Martin P. Hellwig'
AUTHOR_EMAIL = 'martin.hellwig@gmail.com'
URL_MAIN = "https://bitbucket.org/hellwig/" + NAME + '/'
DOWNLOAD_ID = os.environ.get('CI_COMMIT_ID', VERSION)
URL_DOWNLOAD = URL_MAIN + 'get/' + DOWNLOAD_ID + '.zip'

KEYWORDS = [
    'django',
    'django-integrator'
    ]

CLASSIFIERS = [
    'Programming Language :: Python :: 3',
    ]

REQUIREMENTS = [
    'Django',
    'django-integrator',
    ]

LICENSE = 'BSD'

################################################################################

KWARGS = {
    'name':NAME, 'packages':find_packages(), 'version':VERSION,
    'description':DESCRIPTION, 'author':AUTHOR, 'author_email':AUTHOR_EMAIL,
    'url':URL_MAIN, 'download_url':URL_DOWNLOAD, 'keywords':KEYWORDS,
    'license':LICENSE, 'classifiers':CLASSIFIERS,
    'install_requires':REQUIREMENTS}

setup(**KWARGS)
